@@ -1,5 +1,5 @@
 /* Routines for GCC for a Symbian OS targeted SH backend.
-   Copyright (C) 2004, 2005, 2007 Free Software Foundation, Inc.
+   Copyright (C) 2004, 2005, 2007, 2009 Free Software Foundation, Inc.
    Contributed by RedHat.
    Most of this code is stolen from i386/winnt.c.
 
@@ -139,7 +139,7 @@ sh_symbian_dllimport_p (tree decl)
      earlier declaration.  */
   if (TREE_CODE (decl) ==  FUNCTION_DECL
       && DECL_INITIAL (decl)
-      && !DECL_INLINE (decl))
+      && ! DECL_DECLARED_INLINE_P (decl))
     {
       /* Don't warn about artificial methods.  */
       if (!DECL_ARTIFICIAL (decl))
@@ -152,7 +152,7 @@ sh_symbian_dllimport_p (tree decl)
   /* We ignore the dllimport attribute for inline member functions.
      This differs from MSVC behavior which treats it like GNUC
      'extern inline' extension.   */
-  else if (TREE_CODE (decl) == FUNCTION_DECL && DECL_INLINE (decl))
+  else if (TREE_CODE (decl) == FUNCTION_DECL && DECL_DECLARED_INLINE_P (decl))
     {
       if (extra_warnings)
 	warning (OPT_Wattributes, "inline function %q+D is declared as "
@@ -380,7 +380,7 @@ symbian_add_attribute (tree node, const char *attr_name)
     TYPE_ATTRIBUTES (node) = tree_cons (attr, NULL_TREE, attrs);
 
 #if SYMBIAN_DEBUG
-  fprintf (stderr, "propogate %s attribute", attr_name);
+  fprintf (stderr, "propagate %s attribute", attr_name);
   print_node_brief (stderr, " to", node, 0);
   fprintf (stderr, "\n");
 #endif
@@ -580,7 +580,7 @@ symbian_possibly_export_base_class (tree base_class)
 	  if (DECL_PURE_VIRTUAL_P (member))
 	    continue;
 
-	  if (DECL_INLINE (member))
+	  if (DECL_DECLARED_INLINE_P (member))
 	    continue;
 
 	  break;
@@ -671,7 +671,7 @@ symbian_export_vtable_and_rtti_p (tree ctype)
 
 	  if (DECL_CONSTRUCTOR_P (member) || DECL_DESTRUCTOR_P (member))
 	    {
-	      if (DECL_INLINE (member)
+	      if (DECL_DECLARED_INLINE_P (member)
 		  /* Ignore C++ backend created inline ctors/dtors.  */
 		  && (   DECL_MAYBE_IN_CHARGE_CONSTRUCTOR_P (member)
 		      || DECL_MAYBE_IN_CHARGE_DESTRUCTOR_P (member)))
@@ -688,7 +688,7 @@ symbian_export_vtable_and_rtti_p (tree ctype)
 	      if (! DECL_VIRTUAL_P (member))
 		continue;
 
-	      if (DECL_INLINE (member))
+	      if (DECL_DECLARED_INLINE_P (member))
 		continue;
 
 	      if (lookup_attribute ("dllimport", DECL_ATTRIBUTES (member)))
